/*
 * ALSA sequencer MIDI-through client
 * Copyright (c) 1999-2000 by Takashi Iwai <iwai@ww.uni-erlangen.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define SND_MAIN_OBJECT_FILE
#include "../../include/driver.h"
#include "seq_clientmgr.h"

/*

  Sequencer MIDI-through client

  This gives a simple midi-through client.  All the normal input events
  are redirected to output port immediately.
  The routing can be done via aconnect program in alsa-utils.

  Each client has a static client number 62 (= SND_SEQ_CLIENT_DUMMY).
  If you want to auto-load this module, you may add the following alias
  in your /etc/conf.modules file.

	alias snd-seq-client-62  snd-seq-dummy

  The module is loaded on demand for client 62, or /proc/asound/seq/
  is accessed.  If you don't need this module to be loaded, alias
  snd-seq-client-62 as "off".  This will help modprobe.

  The number of ports to be created can be specified via the module
  paramter "ports".  For example, to create four ports, add the
  following option in /etc/modules.conf:

	option snd-seq-dummy ports=4

  The modle option "duplex=1" enables duplex operation to the port.
  In duplex mode, a pair of ports are created instead of single port,
  and events are tunneled between pair-ports.  For example, input to
  port A is sent to output port of another port B and vice versa.
  In duplex mode, each port has DUPLEX capability.

 */


MODULE_PARM(ports, "i");
MODULE_PARM_DESC(ports, "number of ports to be created\n");
MODULE_PARM(duplex, "i");
MODULE_PARM_DESC(duplex, "create DUPLEX ports\n");
int ports = 1;
int duplex = 0;

typedef struct port_info {
	int client;
	int port;
	int duplex;
	int connect;
} port_info_t;

static int my_client = -1;

/*
 * unuse callback - send ALL_SOUNDS_OFF events to subscribers
 * Note: this callback is called only after all subscribers are removed.
 */
static int
dummy_unuse(void *private_data, snd_seq_port_subscribe_t *info)
{
	port_info_t *p = private_data;
	int i;
	snd_seq_event_t ev;

	memset(&ev, 0, sizeof(ev));
	if (p->duplex)
		ev.source.port = p->connect;
	else
		ev.source.port = p->port;
	ev.dest.client = SND_SEQ_ADDRESS_SUBSCRIBERS;
	ev.type = SND_SEQ_EVENT_CONTROLLER;
	ev.data.control.param = SND_MCTL_ALL_SOUNDS_OFF;
	for (i = 0; i < 16; i++) {
		ev.data.control.channel = i;
		snd_seq_kernel_client_dispatch(p->client, &ev, 0, 0);
	}
	return 0;
}

/*
 * event input callback - just redirect events to subscribers
 */
static int
dummy_input(snd_seq_event_t *ev, int direct, void *private_data, int atomic, int hop)
{
	port_info_t *p = private_data;

	if (ev->source.client == SND_SEQ_CLIENT_SYSTEM)
		return 0; /* ignore system messages */
	if (p->duplex)
		ev->source.port = p->connect;
	else
		ev->source.port = p->port;
	ev->dest.client = SND_SEQ_ADDRESS_SUBSCRIBERS;
	return snd_seq_kernel_client_dispatch(p->client, ev, atomic, hop);
}

/*
 * free_private callback
 */
static void
dummy_free(void *private_data)
{
	snd_kfree(private_data);
}

/*
 * create a port
 */
static port_info_t __init *
create_port(int idx, int type)
{
	snd_seq_port_info_t pinfo;
	snd_seq_port_callback_t pcb;
	port_info_t *rec;

	if ((rec = snd_kmalloc(sizeof(*rec), GFP_KERNEL)) == NULL)
		return NULL;

	rec->client = my_client;
	rec->duplex = duplex;
	rec->connect = 0;
	memset(&pinfo, 0, sizeof(pinfo));
	pinfo.client = my_client;
	if (duplex)
		sprintf(pinfo.name, "Midi Through Port-%d:%c", idx,
			(type ? 'B' : 'A'));
	else
		sprintf(pinfo.name, "Midi Through Port-%d", idx);
	pinfo.capability = SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ;
	pinfo.capability |= SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE;
	if (duplex)
		pinfo.capability |= SND_SEQ_PORT_CAP_DUPLEX;
	pinfo.type = SND_SEQ_PORT_TYPE_MIDI_GENERIC;
	memset(&pcb, 0, sizeof(pcb));
	pcb.unuse = dummy_unuse;
	pcb.event_input = dummy_input;
	pcb.private_free = dummy_free;
	pcb.private_data = rec;
	pinfo.kernel = &pcb;
	if (snd_seq_kernel_client_ctl(my_client, SND_SEQ_IOCTL_CREATE_PORT, &pinfo) < 0) {
		snd_kfree(rec);
		return NULL;
	}
	rec->port = pinfo.port;
	return rec;
}

/*
 * register client and create ports
 */
static int __init
register_client(void)
{
	snd_seq_client_callback_t cb;
	snd_seq_client_info_t cinfo;
	port_info_t *rec1, *rec2;
	int i;

	if (ports < 1) {
		snd_printk("seq_dummy: invalid number of ports %d\n", ports);
		return -EINVAL;
	}

	/* create client */
	memset(&cb, 0, sizeof(cb));
	cb.allow_input = 1;
	cb.allow_output = 1;
	my_client = snd_seq_create_kernel_client(NULL, SND_SEQ_CLIENT_DUMMY, &cb);
	if (my_client < 0)
		return my_client;

	/* set client name */
	memset(&cinfo, 0, sizeof(cinfo));
	cinfo.client = my_client;
	cinfo.type = KERNEL_CLIENT;
	strcpy(cinfo.name, "Midi Through");
	strcpy(cinfo.group, SND_SEQ_GROUP_DEVICE);
	snd_seq_kernel_client_ctl(my_client, SND_SEQ_IOCTL_SET_CLIENT_INFO, &cinfo);

	/* create ports */
	for (i = 0; i < ports; i++) {
		rec1 = create_port(i, 0);
		if (rec1 == NULL) {
			snd_seq_delete_kernel_client(my_client);
			return -ENOMEM;
		}
		if (duplex) {
			rec2 = create_port(i, 1);
			if (rec2 == NULL) {
				snd_seq_delete_kernel_client(my_client);
				return -ENOMEM;
			}
			rec1->connect = rec2->port;
			rec2->connect = rec1->port;
		}
	}

	return 0;
}

#ifdef MODULE
/*
 * delete client if exists
 */
static void __exit
delete_client(void)
{
	if (my_client >= 0)
		snd_seq_delete_kernel_client(my_client);
}
#endif


/*
 *  Init part
 */

#ifdef MODULE
int __init init_module(void)
#else
int alsa_seq_dummy_init(void)
#endif
{
	return register_client();
}

#ifdef MODULE

void __exit cleanup_module(void)
{
	delete_client();
}

#endif

MODULE_AUTHOR("Takashi Iwai <iwai@ww.uni-erlangen.de>");
MODULE_DESCRIPTION("ALSA sequencer MIDI-through client");
MODULE_SUPPORTED_DEVICE("sound");

